/**************************************************************************************************

Copyright (c) 2015   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mtk_dome_master.h"
#include  <stdio.h>
#include  "../common/math/vector3.h"
#include  "../common/math/quaternion.h"
#include  "mtk_camera_setting.h"
#include  "mtk_draw_struct.h"
#include  "mtk_mode.h"
#include  "mtk_objects.h"
#include  "mtk_planetarium.h"



//--  Externs
void  draw_scene_all(const mtk_draw_struct& ds, bool bClear);
mtk_draw_struct calc_draw_struct(const mtk_camera& camera_base, float aspc_rt, int _vp_x1, int _vp_x2, int _vp_y1, int _vp_y2, vector3f dr, bool bRotBB, float line_w);



//
static int buffer_size  = 1024;		// FBÕTCY
static int capture_size = 2048;		// Lv`[摜̃TCY
static int mesh_N = 50;				// bV̕
static float elev_angle = 30.0;		// ʕ̋p
static float label_mag = 2.0;		// x̊g嗦
static float line_width_fac = 2.5;	// 


void
set_dome_master_buffer_size(int sz)
{
	if (sz < 128)	{ sz = 128; }
	if (sz > 4096)	{ sz = 4096; }

	buffer_size = sz;
}

void
set_dome_master_capture_size(int sz)
{
	if (sz < 100)	{ sz = 100; }
	if (sz > 10000)	{ sz = 10000; }

	capture_size = sz;
}

void
set_dome_master_mesh_N(int N)
{
	if (N < 4)		{ N = 4; }
	if (N > 2000)	{ N = 2000; }

	mesh_N = N;
}

void
set_dome_master_elev_angle(float angle)
{
	elev_angle = angle;
}

float
get_dome_master_elev_angle()
{
	return elev_angle;
}

void
set_dome_master_label_mag(float fac)
{
	label_mag = fac;
}

float
get_dome_master_label_mag()
{
	return label_mag;
}

void
set_dome_master_line_width_fac(float fac)
{
	line_width_fac = fac;
}

float
get_dome_master_line_width_fac()
{
	return line_width_fac;
}




void
check_extentions()
{
	char ext_name[][50] = {
		"GL_ARB_framebuffer_object",
		"GL_EXT_framebuffer_object"
	};
	const int N = sizeof(ext_name) / sizeof(ext_name[0]);

	FILE *fp;


	fp = fopen("OpenGL_ext_check.txt", "wt");
	for (int i = 0; i < N; i++) {
		char str[2][20] = {"NG", "OK"};
		bool bSupport;
		int j;

		bSupport = glewIsExtensionSupported(ext_name[i]);
		j = (bSupport) ? 1 : 0;

		fprintf(fp, "%s %s\n", ext_name[i], str[j]);
	}
	fclose(fp);
}






//---------------
//  Dome master
//---------------
const vector3f _eF[6] = {
	vector3f(1, 0, 0), vector3f(-1, 0, 0),
	vector3f(0, 1, 0), vector3f(0, -1, 0),
	vector3f(0, 0, 1), vector3f(0, 0, -1)
};
const vector3f _eU[6] = {
	vector3f(0, 0, 1), vector3f(0, 0, 1),
	vector3f(0, 0, 1), vector3f(0, 0, 1),
	vector3f(0, -1, 0), vector3f(0, -1, 0)
};



class gl_dome_master
{
	int N, _Nx[5], _Ny[5], _idx0[5];
	float *vtxbuf, *uvbuf;
	unsigned int *idxbuf;

public:
	gl_dome_master(const int _N);
	~gl_dome_master();

	void	Draw(int face) const;
};


gl_dome_master::gl_dome_master(const int _N) : vtxbuf(0), uvbuf(0), idxbuf(0)
{
	N = _N;
	vtxbuf = new float[5 * N*N * 3];
	uvbuf = new float[5 * N*N * 2];
	idxbuf = new unsigned int[5 * N * N * 2];

	int idx0 = 0;
	int addr = 0;
	
	for (int face = 0; face < 5; face++) {

		vector3f eF, eU, eR;
		eF = _eF[face];
		eU = _eU[face];
		eR = normalize(cross_product(eU, -eF));


		int ix1, ix2, Nx;
		float u0, u1;
		ix1 = 0;
		ix2 = N;
		u0 = 0;
		u1 = 1;
		Nx = ix2 - ix1;

		int iy1, iy2, Ny;
		float v0, v1;
		iy1 = 0;
		iy2 = N;
		v0 = 0;
		v1 = 1;
		if (face != 4) { iy1 = N / 2; iy2 = N; v0 = 0.5; }	// using half area
		Ny = iy2 - iy1;

		_Nx[face] = Nx;
		_Ny[face] = Ny;

		// make vertex
		for (int iy = 0; iy < Ny; iy++) {
			float ry, v, _y;

			ry = float(iy) / (Ny - 1);
			v = (v1 - v0) * ry + v0;;
			_y = 2.0 * v - 1.0;


			for (int ix = 0; ix < Nx; ix++) {
				float rx, u, _x;

				rx = float(ix) / (Nx - 1);
				u = (u1 - u0) * rx + u0;
				_x = 2.0 * u - 1.0;


				vector3f pos = _x*eR + _y*eU + 1.0f*eF;


				float x, y, z, rho, th, phi, r;
				x = pos.x;
				y = pos.y;
				z = pos.z;
				rho = sqrt(x*x + y*y);
				th = atan2(rho, z);
				phi = atan2(y, x);
				r = th / (0.5 * M_PI);

				// position on dome master
				float xx, yy;
				xx = r * cos(phi);
				yy = r * sin(phi);


				// vertex position
				int idx = (idx0 + iy*Nx + ix) * 3;
				vtxbuf[idx] = xx;
				vtxbuf[idx + 1] = yy;
				vtxbuf[idx + 2] = 0.0;

				// tex coord
				idx = (idx0 + iy*Nx + ix) * 2;
				uvbuf[idx] = u;
				uvbuf[idx + 1] = v;
			}
		}

		// make index
		_idx0[face] = addr;
		for (int iy = 0; iy < Ny; iy++) {
			for (int ix = 0; ix < Nx; ix++) {
				int idx;

				idx = idx0 + iy*Nx + ix;
				idxbuf[addr++] = idx;	// to be counter-clock wise
				idxbuf[addr++] = idx + Nx;
			}
		}

		idx0 += Nx*Ny;
	}
}


gl_dome_master::~gl_dome_master()
{
	delete[] vtxbuf;
	delete[] uvbuf;
	delete[] idxbuf;
}


void
gl_dome_master::Draw(int face) const
{
	int Nx, Ny, idx0;
	Nx = _Nx[face];
	Ny = _Ny[face];
	idx0 = _idx0[face];

	for (int i = 0; i < Ny - 1; i++) {
		glBegin(GL_TRIANGLE_STRIP);
		for (int j = 0; j < Nx * 2; j++) {
			int idx;

			idx = idxbuf[idx0 + j + Nx*2*i];

			// tex coord
			glTexCoord2f(uvbuf[2*idx], uvbuf[2*idx + 1]);

			// vertex position
			glVertex2f(vtxbuf[3*idx], vtxbuf[3*idx + 1]);
		}
		glEnd();
	}
}


void
draw_dome_master(const mtk_camera camera_base, int w, int h, gl_frame_buffer *fbo_out)
{
	static gl_dome_master dm(mesh_N);
	static gl_frame_buffer fbo;
	static bool first = true;

	if (first) {
		//check_extentions();

		glEnable(GL_TEXTURE_2D);
		fbo.Init(buffer_size, buffer_size, true);
		first = false;
	}

	int fbo_w, fbo_h;
	fbo_w = fbo.get_width();
	fbo_h = fbo.get_height();

	mtk_draw_struct  ds[5];
	quaternionf _ori[5];
	for (int face = 0; face < 5; face++) {
		vector3f eF, eU;
		quaternionf ori;

		eF = _eF[face];
		eU = _eU[face];
		ori = axes_to_quaternion(eF, eU);

		ori = make_rotation_quaternion(float(M_PI), vector3f(1, 0, 0)) * ori;
		

		// Evevation angle
		ori = make_rotation_quaternion(float(D2R * (90.0-elev_angle)), vector3f(1, 0, 0)) * ori;


		_ori[face] = ori;
		set_local_camera(ori, vector3d(0, 0, 0));


		// set draw_struct and frastum
		bool bRotBB = true;

		float z1, z2;
		z1 = get_z_near();
		z2 = get_z_mid();

		if (face == 4) {
			ds[face] = calc_draw_struct(camera_base, 1.0, 0, fbo_w, 0, fbo_h, vector3f(0, 0, 0), bRotBB, line_width_fac);
			ds[face].frs_near.set(-z1, z1, -z1, z1, z1, z2);
			ds[face].frs_far.set(-z2, z2, -z2, z2, z2, get_z_far());
		}
		else {
			ds[face] = calc_draw_struct(camera_base, 1.0, 0, fbo_w, fbo_h / 2, fbo_h, vector3f(0, 0, 0), bRotBB, line_width_fac);
			ds[face].frs_near.set(-z1, z1, 0, z1, z1, z2);
			ds[face].frs_far.set(-z2, z2, 0, z2, z2, get_z_far());
		}
		ds[face].set_red_per_pixel();
	}


	float x, y;
	if (w > h) {
		x = float(w) / h;
		y = 1;
	}
	else {
		x = 1;
		y = float(h) / w;
	}


	if (fbo_out != 0) {
		fbo_out->Bind();
	}
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	if (fbo_out != 0) {
		fbo_out->Unbind();
	}


	for (int face = 0; face < 5; face++) {
		
		//---  draw on FBO  ---
		fbo.Bind();
		glViewport(0, 0, fbo_w, fbo_h);


		// draw scene
		//***
		set_local_camera(_ori[face], vector3d(0, 0, 0));
		//***

		if (ds[face].tyMode == MODE_3D) {
			prepare_draw_objects(&ds[face], 1);
		}
		else {
			prepare_draw_planetarium(&ds[face], 1);
		}


		draw_scene_all(ds[face], true);


		fbo.Unbind();
		

		
		//---  draw on dome master  ---
		if (fbo_out != 0) {
			fbo_out->Bind();
		}

		glColor3f(1, 1, 1);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_CULL_FACE);
		glDisable(GL_BLEND);
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(-x, x, y, -y, -1, 1);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		glViewport(0, 0, w, h);
		glBindTexture(GL_TEXTURE_2D, fbo.get_texture());
		dm.Draw(face);

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		if (fbo_out != 0) {
			fbo_out->Unbind();
		}
	}


	//**
	// ̕\[hɐ؂ւƂ̂߂ɁAoriPʃNH[^jIɖ߂
	set_local_camera(quaternionf(1,0,0,0), vector3d(0, 0, 0));
}



void
capture_dome_master(const mtk_camera camera_base, image& img)
{
	gl_frame_buffer fbo_out;

	fbo_out.Init(capture_size, capture_size, false);

	draw_dome_master(camera_base, capture_size, capture_size, &fbo_out);

	fbo_out.Bind();
	gl_read_pixels(img, 0, 0, capture_size, capture_size);
	fbo_out.Unbind();
}


